from sklearn.preprocessing import LabelEncoder
from sklearn.preprocessing import OneHotEncoder
import numpy as np
from mavenn.src.error_handling import handle_errors, check
import matplotlib.pyplot as plt
import logomaker
import seaborn as sns
import pandas as pd
import mavenn

@handle_errors
def onehot_sequence(sequence, bases):

    """
    Encodes a single sequence into a one-hot matrix

    parameters
    ----------

    sequence: (str)
        string that needs to be turned into a one-hot encoded matrix

    bases: (list)
        specifies unique characters in the sequence


    returns
    -------
    oh_encoded_vector: (np.array)
        one-hot encoded array for the input sequence.


    """

    # sklearn objects and operations need for one-hot encoding
    label_encoder = LabelEncoder()
    label_encoder.fit(bases)
    tmp = label_encoder.transform(bases)
    tmp = tmp.reshape(len(tmp), 1)
    onehot_encoder = OneHotEncoder(sparse=False)
    onehot_encoder.fit(tmp)

    # perform one-hot encoding:
    categorical_vector = label_encoder.transform(list(sequence))

    # reshape so that array has correct dimensions for input into tf.
    categorical_vector = categorical_vector.reshape(len(categorical_vector), 1)
    oh_encoded_vector = onehot_encoder.transform(categorical_vector)
    return oh_encoded_vector


@handle_errors
def onehot_encode_array(data, bases_dict, ohe_single_batch_size=10000):

    """
    one-hot encode sequences in batches in a vectorized way

    parameters
    ----------

    data: (array-like)
        data which will be one-hot encoded

    bases_dict: (str)
        Specifies the type of input sequences. Three possible choices
        allowed: ['dna','rna','protein']

    ohe_single_batch_size: (int)
        integer specifying how many sequences to one-hot encode at a time.
        The larger this number number, the quicker the encoding will happen,
        but this may also take up a lot of memory and throw an exception
        if its too large.

    returns
    -------
    input_seqs_ohe: (np array)
        array of one-hot encoded sequences based on the input data


    """

    # validate that sequences is array=like
    check(isinstance(data, (tuple, list, np.ndarray)),
          'type(data) = %s; data must be array-like.' %
          type(data))

    # check that ohe_single_batch_size is an integer
    check(isinstance(ohe_single_batch_size, (int, np.int64)),
          'type(ohe_single_batch_size) = %s must be of type int or numpy.int64' % type(ohe_single_batch_size))

    # check that data[0] passed has length, which will be used to get sequence length
    check(hasattr(data[0], '__len__'), 'entered sequence data does not have length, needs to have length > 0')

    sequence_length = len(data[0])

    # container list for batches of oh-encoded sequences
    input_seqs_ohe_batches = []

    # partitions of batches
    ohe_batches = np.arange(0, len(data), ohe_single_batch_size)
    for ohe_batch_index in range(len(ohe_batches)):
        if ohe_batch_index == len(ohe_batches) - 1:
            # OHE remaining sequences (that are smaller than batch size)
            input_seqs_ohe_batches.append(
                onehot_sequence(''.join(data[ohe_batches[ohe_batch_index]:]), bases=bases_dict)
                    .reshape(-1, sequence_length, len(bases_dict)))
        else:
            # OHE sequences in batches
            input_seqs_ohe_batches.append(onehot_sequence(
                ''.join(data[ohe_batches[ohe_batch_index]:ohe_batches[ohe_batch_index + 1]]), bases=bases_dict)
                                          .reshape(-1, sequence_length, len(bases_dict)))

    # this array will contain the one-hot encoded sequences
    input_seqs_ohe = np.array([])

    # concatenate all the oh-encoded batches
    for batch_index in range(len(input_seqs_ohe_batches)):
        input_seqs_ohe = np.concatenate([input_seqs_ohe, input_seqs_ohe_batches[batch_index]
                                             .ravel()]).copy()

    # reshape so that shape of oh-encoded array is [number samples, sequence_length*alphabet_dict]
    input_seqs_ohe = input_seqs_ohe.reshape(len(data), sequence_length * len(bases_dict)).copy()

    return input_seqs_ohe


@handle_errors
def _generate_nbr_features_from_sequences(sequences,
                                          alphabet_dict='dna'):

    """
    Method that takes in sequences are generates sequences
    with neighbor features

    parameters
    ----------

    sequences: (array-like)
        array contains raw input sequences

    alphabet_dict: (str)
        Specifies the type of input sequences. Three possible choices
        allowed: ['dna','rna','protein']

    returns
    -------
    nbr_sequences: (array-like)
        Data Frame of sequences where each row contains a sequence example
        with neighbor features

    """

    # validate that sequences is array=like
    check(isinstance(sequences, (tuple, list, np.ndarray)),
          'type(sequences) = %s; sequences must be array-like.' %
          type(sequences))

    # check that alphabet_dict is valid
    check(alphabet_dict in {'dna', 'rna', 'protein'},
          'alphabet_dict = %s; must be "dna", "rna", or "protein"' %
          alphabet_dict)

    if alphabet_dict == 'dna':
        bases = ['A', 'C', 'G', 'T']
    elif alphabet_dict == 'rna':
        bases = ['A', 'C', 'G', 'U']
    elif alphabet_dict == 'protein':

        # this should be called amino-acids
        # need to figure out way to deal with
        # naming without changing a bunch of
        # unnecessary refactoring.
        bases = ['A', 'C', 'D', 'E', 'F',
                 'G', 'H', 'I', 'K', 'L',
                 'M', 'N', 'P', 'Q', 'R',
                 'S', 'T', 'V', 'W', 'Y']

    # form neighbor dinucleotide features that will
    # be used to one-hot encode sequnces
    nbr_dinucleotides = []

    for i in range(len(bases)):
        for j in range(len(bases)):
            nbr_dinucleotides.append(bases[i] + bases[j])

    # one-hot encode di-nucleotide training set
    dinuc_seqs_OHE = []
    for _ in range(len(sequences)):
        # take current raw training sequence
        raw_sequence = sequences[_]

        # split it into di-nucleotide pairs
        di_nucl_pairs = [raw_sequence[i:i + 2] for i in range(0, len(raw_sequence) - 1, 1)]

        # get indices of where pairs occur so that these indices could be used to one-hot encode.
        list_of_nbr_indices = [nbr_dinucleotides.index(dn) for dn in di_nucl_pairs]

        # do One-hot encoding. Every time a pair from list 'nbr_dinucleotides'
        # appears at a position, put 1 there, otherwise zeros.
        tmp_seq = np.array(list_of_nbr_indices)
        OHE_dinucl_seq = np.zeros((tmp_seq.size, len(nbr_dinucleotides)))
        OHE_dinucl_seq[np.arange(tmp_seq.size), tmp_seq] = 1

        dinuc_seqs_OHE.append(OHE_dinucl_seq.ravel())

    return np.array(dinuc_seqs_OHE)


def _generate_all_pair_features_from_sequences(sequences,
                                               alphabet_dict='dna'):

    """
    Method that takes in sequences are generates sequences
    with all pair features

    parameters
    ----------

    sequences: (array-like)
        array contains raw input sequences

    alphabet_dict: (str)
        Specifies the type of input sequences. Three possible choices
     allowed: ['dna','rna','protein']


    returns
    -------

    all_pairs_sequences: (array-like)
        Data Frame of sequences where each row contains a sequence example
        with all-pair features

    """

    # validate that sequences is array=like
    check(isinstance(sequences, (tuple, list, np.ndarray)),
          'type(sequences) = %s; sequences must be array-like.' %
          type(sequences))

    # check that alphabet_dict is valid
    check(alphabet_dict in {'dna', 'rna', 'protein'},
          'alphabet_dict = %s; must be "dna", "rna", or "protein"' %
          alphabet_dict)

    if alphabet_dict == 'dna':
        bases = ['A', 'C', 'G', 'T']
    elif alphabet_dict == 'rna':
        bases = ['A', 'C', 'G', 'U']
    elif alphabet_dict == 'protein':

        # this should be called amino-acids
        # need to figure out way to deal with
        # naming without changing a bunch of
        # unnecessary refactoring.
        bases = ['A', 'C', 'D', 'E', 'F',
                 'G', 'H', 'I', 'K', 'L',
                 'M', 'N', 'P', 'Q', 'R',
                 'S', 'T', 'V', 'W', 'Y']

    # form neighbor dinucleotide features that will
    # be used to one-hot encode sequnces
    allpair_dinucleotides = []

    for i in range(len(bases)):
        for j in range(len(bases)):
            allpair_dinucleotides.append(bases[i] + bases[j])

    # one-hot encode di-nucleotide training set
    allpairs_seqs_OHE = []
    for _ in range(len(sequences)):

        # take current raw training sequence
        raw_sequence = sequences[_]

        # split it into all nucleotide pairs
        all_nucl_pairs = []

        for i in range(len(raw_sequence)):
            for j in range(i + 1, len(raw_sequence)):
                all_nucl_pairs.append(raw_sequence[i] + raw_sequence[j])

        # get indices of where pairs occur so that these indices could be used to one-hot encode.
        list_of_allpair_indices = [allpair_dinucleotides.index(dn) for dn in all_nucl_pairs]

        # do One-hot encoding. Every time a pair from list 'allpair_dinucleotides'
        # appears at a position, put 1 there, otherwise zeros.
        tmp_seq = np.array(list_of_allpair_indices)
        OHE_dinucl_seq = np.zeros((tmp_seq.size, len(allpair_dinucleotides)))
        OHE_dinucl_seq[np.arange(tmp_seq.size), tmp_seq] = 1

        allpairs_seqs_OHE.append(OHE_dinucl_seq.ravel())

    return np.array(allpairs_seqs_OHE)


@handle_errors
def ge_plots_for_mavenn_demo(loss_history,
                             predictions,
                             true_labels,
                             sequences,
                             GE_model):
    """

    Helper function that plots predictions vs true labels,
    losses and the GE nonlinearity for mavenn's demo functions

    parameters
    ----------
    loss_history: (tf loss history/array-like)
        arrays containing loss values

    sequences: (array-like)
        sequences which will be used to make predictions

    predictions: (array-like)
        model predictions for the sequences entered

    true_labels: (array-like)
        true labels for the model predictions

    GE_model: (GlobalEpistasisModel object)
        trained GE model from which the GE nonlinearity will be plotted


    returns
    -------
    None

    """

    fig, ax = plt.subplots(1, 3, figsize=(10, 3))

    ax[0].plot(loss_history.history['loss'], color='blue')
    ax[0].plot(loss_history.history['val_loss'], color='orange')
    ax[0].set_title('Model loss', fontsize=12)
    ax[0].set_ylabel('loss', fontsize=12)
    ax[0].set_xlabel('epoch', fontsize=12)
    ax[0].legend(['train', 'validation'])

    ax[1].scatter(predictions, true_labels, s=1, alpha=0.5, color='black')
    ax[1].set_ylabel('observations')
    ax[1].set_xlabel('predictions')

    # get ge nonlinear function
    GE_nonlinearity = GE_model.ge_nonlinearity(sequences)

    ax[2].plot(GE_nonlinearity[1], GE_nonlinearity[0], color='black')
    ax[2].scatter(GE_nonlinearity[2], true_labels, color='gray', s=1, alpha=0.5)

    ax[2].set_ylabel('observations')
    ax[2].set_xlabel('latent trait ($\phi$)')

    plt.tight_layout()
    plt.show()


@handle_errors
def na_plots_for_mavenn_demo(loss_history,
                             NAR,
                             noise_model,
                             phi_range):
    """

    Helper function that plots inferred additive parameters,
    losses and the NA noise model for mavenn's demo functions

    parameters
    ----------
    loss_history: (tf loss history/array-like)
        arrays containing loss values

    NAR: (NoiseAgnosticModel object)
        trained NA model from which additive parameters
        will be extracted and plotted as a sequence logo

    noise_model: (array-like)
        the inferred noise model which will be plotted
        as a heatmap


    returns
    -------
    None
    """

    # plot results
    fig, ax = plt.subplots(1, 3, figsize=(10, 3))

    ax[0].plot(loss_history.history['loss'], color='blue')
    ax[0].plot(loss_history.history['val_loss'], color='orange')
    ax[0].set_title('Model loss')
    ax[0].set_ylabel('loss')
    ax[0].set_xlabel('epoch')
    ax[0].legend(['train', 'validation'])

    # make logo to visualize additive parameters
    ax[1].set_ylabel('additive parameters')
    ax[1].set_xlabel('position')

    #theta_df = pd.DataFrame(NAR.nn_model().layers[1].get_weights()[0].reshape(39, 4),columns=['A', 'C', 'G', 'T'])
    theta_df = pd.DataFrame(NAR.return_theta().reshape(39, 4), columns=['A', 'C', 'G', 'T'])


    additive_logo = logomaker.Logo(theta_df,
                                   center_values=True,
                                   ax=ax[1])

    # view the inferred noise model as a heatmap
    # noise_model_heatmap = sns.heatmap(noise_model.T, cmap='Greens', ax=ax[2])
    # ax[2].invert_yaxis()
    # ax[2].set_xticks(([0,int(len(phi_range)/2), len(phi_range)-2]), minor=False)
    # ax[2].set_xticklabels(([str(phi_range[0]), 0, str(phi_range[len(phi_range)-1])]), minor=False)
    # ax[2].set_ylabel(' bin numbers')
    # ax[2].set_xlabel('latent trait ($\phi$)')

    if noise_model.T[noise_model.T.shape[0] - 1][0] > noise_model.T[noise_model.T.shape[0] - 1][
                noise_model.T.shape[1] - 1]:
        noise_model_heatmap = sns.heatmap(pd.DataFrame(noise_model.T).loc[::1, ::-1], cmap='Greens', ax=ax[2])
    else:
        noise_model_heatmap = sns.heatmap(noise_model.T, cmap='Greens', ax=ax[2])
    ax[2].invert_yaxis()
    ax[2].set_xticks(([0, int(len(phi_range) / 2), len(phi_range) - 2]), minor=False)
    middle_tick = str(phi_range[int(len(phi_range) / 2)])
    ax[2].set_xticklabels(([str(phi_range[0])[1:5], middle_tick[1:5], str(phi_range[len(phi_range) - 1])[1:5]]),
                       minor=False)
    ax[2].set_ylabel(' bin numbers')
    ax[2].set_xlabel('latent trait ($\phi$)')


    plt.tight_layout()
    plt.show()

@handle_errors
def load_olson_data_GB1():

    """
    Helper function to turn data provided by Olson et al.
    into sequence-values arrays. This method is used in the
    GB1 GE demo.


    return
    ------
    gb1_df: (pd dataframe)
        dataframe containing sequences (single and double mutants)
        and their corresponding log enrichment values

    """

    # GB1 WT sequences
    WT_seq = 'QYKLILNGKTLKGETTTEAVDAATAEKVFKQYANDNGVDGEWTYDDATKTFTVTE'

    # WT sequence library and selection counts.
    WT_input_count = 1759616
    WT_selection_count = 3041819

    # load double mutant data
    oslon_mutant_positions_data = pd.read_csv(mavenn.__path__[0] +
                                              '/examples/datafiles/gb1/oslon_data_double_mutants_ambler.csv',
                                              na_values="nan")

    # lists that will contain sequences and their values
    sequences = []
    enrichment = []

    for loop_index in range(len(oslon_mutant_positions_data)):

        # skip row 259455 containing, it contains bad data
        if loop_index == 259455:
            continue

        # get positions of double mutants in sequence
        mut_1_index = int(oslon_mutant_positions_data['Mut1 Position'][loop_index]) - 2
        mut_2_index = int(oslon_mutant_positions_data['Mut2 Position'][loop_index]) - 2

        # get identity of mutations
        mut_1 = oslon_mutant_positions_data['Mut1 Mutation'][loop_index]
        mut_2 = oslon_mutant_positions_data['Mut2 Mutation'][loop_index]

        # form full mutant sequence.
        temp_dbl_mutant_seq = list(WT_seq)
        temp_dbl_mutant_seq[mut_1_index] = mut_1
        temp_dbl_mutant_seq[mut_2_index] = mut_2

        if loop_index % 100000 == 0:
            print('generating data: %d out of %d' % (loop_index,len(oslon_mutant_positions_data)))

        # calculate enrichment for double mutant sequence sequence
        input_count = oslon_mutant_positions_data['Input Count'][loop_index]
        selection_count = oslon_mutant_positions_data['Selection Count'][loop_index]
        # added pseudocount to ensure log doesn't throw up
        temp_fitness = ((selection_count + 1) / input_count) / (WT_selection_count / WT_input_count)

        # append sequence
        sequences.append(''.join(temp_dbl_mutant_seq))
        enrichment.append(temp_fitness)

    # load single mutants data
    oslon_single_mutant_positions_data = pd.read_csv(mavenn.__path__[0] +
                                                     '/examples/datafiles/gb1/oslon_data_single_mutants_ambler.csv',
                                                     na_values="nan")

    for loop_index in range(len(oslon_single_mutant_positions_data)):
        mut_index = int(oslon_single_mutant_positions_data['Position'][loop_index]) - 2

        mut = oslon_single_mutant_positions_data['Mutation'][loop_index]

        temp_seq = list(WT_seq)
        temp_seq[mut_index] = mut

        # calculate enrichment for sequence
        input_count = oslon_single_mutant_positions_data['Input Count'][loop_index]
        selection_count = oslon_single_mutant_positions_data['Selection Count'][loop_index]
        # added pseudo count to ensure log doesn't throw up
        temp_fitness = ((selection_count + 1) / input_count) / (WT_selection_count / WT_input_count)

        sequences.append(''.join(temp_seq))
        enrichment.append(temp_fitness)

    enrichment = np.array(enrichment).copy()

    gb1_df = pd.DataFrame({'sequence': sequences, 'values': np.log(enrichment)}, columns=['sequence', 'values'])
    return gb1_df


@handle_errors
def get_example_dataset(name='MPSA'):
    """

    Load example sequence-function datasets that
    come with the mavenn package.

    Parameters:
    -----------

    name: (str)
        Name of example dataset. Must be one of
        ('MPSA', 'Sort-Seq', 'GB1-DMS')

    Returns:
    --------
    X, y: (array-like)
        An array containing sequences X and an
        array containing their target values y
    """

    # check that parameter 'name' is valid
    check(name in {'MPSA', 'Sort-Seq', 'GB1-DMS'},
          'name = %s; must be "MPSA", "Sort-Seq", or "GB1-DMS"' %name)

    if name == 'MPSA':

        mpsa_df = pd.read_csv(mavenn.__path__[0] + '/examples/datafiles/mpsa/psi_9nt_mavenn.csv')
        mpsa_df = mpsa_df.dropna()
        mpsa_df = mpsa_df[mpsa_df['values'] > 0]  # No pseudocounts

        return mpsa_df['sequence'].values, np.log10(mpsa_df['values'].values)

    elif name == 'Sort-Seq':

        sequences = np.loadtxt(mavenn.__path__[0] + '/examples/datafiles/sort_seq/full-wt/rnap_sequences.txt',
                               dtype='str')
        bin_counts = np.loadtxt(mavenn.__path__[0] + '/examples/datafiles/sort_seq/full-wt/bin_counts.txt')

        return sequences, bin_counts

    elif name == 'GB1-DMS':

        gb1_df = load_olson_data_GB1()
        return gb1_df['sequence'].values, gb1_df['values'].values


@handle_errors
def _center_matrix(df):
    """
    Centers each row of a matrix about zero by subtracting out the mean.
    This method is used for gauge-fixing additive latent models

    parameters
    ----------

    df: (pd.DataFrame)
        a (L x C) pandas dataframe which contains additive latent model
        parameters

    returns
    -------
    out_df: (pd.DataFrame)
        dataframe that is mean centered.

    theta_bar: (array-like)
        mean of parameters at position l

    """

    # Compute mean value of each row
    means = df.mean(axis=1).values
    theta_bar = -df.mean(axis=1).values[:, np.newaxis]

    # Subtract out means
    out_df = df.copy()
    out_df.loc[:, :] = df.values - means[:, np.newaxis]

    return out_df, theta_bar

@handle_errors
def fix_gauge_neighbor_model(sequenceLength,
                             alphabetSize,
                             theta):
    """Calculates the hierarchical gauge that prioritizes the
    constant feature over neighbor features and the neighbor
    features' coefficients are in the canonical gauge.

    parameters
    ----------
    sequenceLength: (int)
        length of one input sequence

    alphabetSize: (int)
        number of unique bases or amino-acids; e.g. 4 for DNA and 20 for proteins

    theta: (array-like)
        the parameters of the latent neigbhor model

    returns
    -------
    thetaGaugeFixed: (array-like)
        the gauge fixed neighbor parameters

    centeringConstant: (array-like)
        the additive shift relating phi to phi''

    """

    theta = theta.reshape(sequenceLength - 1, alphabetSize, alphabetSize)

    # center parameter values corresponding to a fixed first index, i.e. fixed neighboring position pair:
    centeringConstant = sum(theta.mean(axis=(1, 2)))
    thetaCentered = theta - theta.mean(axis=(1, 2), keepdims=True)

    # calculate canonical gauge for neighbor features:
    thetaGaugeFixed = thetaCentered
    if sequenceLength > 2:
        sRow = np.sum(thetaCentered, axis=2,
                      keepdims=True)  # define array that broadcasts with theta and contains theta's row sums
        sCol = np.sum(thetaCentered, axis=1,
                      keepdims=True)  # define array that broadcasts with theta and contains theta's column sums
        pos = 0
        thetaGaugeFixed[pos] = thetaCentered[pos] + 1 / (2 * alphabetSize) * (-sCol[pos] + sRow[pos + 1].transpose())
        for pos in range(1, sequenceLength - 2):
            thetaGaugeFixed[pos] = thetaCentered[pos] + 1 / (2 * alphabetSize) * (
            sCol[pos - 1].transpose() - sRow[pos] - sCol[pos] + sRow[pos + 1].transpose())
        pos = sequenceLength - 2
        thetaGaugeFixed[pos] = thetaCentered[pos] + 1 / (2 * alphabetSize) * (sCol[pos - 1].transpose() - sRow[pos])
    thetaGaugeFixed = thetaGaugeFixed.ravel()

    return thetaGaugeFixed, centeringConstant



@handle_errors
def kronecker_delta(n1, n2):
    """

    helper method implementing Kronecker delta,
    used in fix_gauge_pairwise_model
    """

    if n1 == n2:
        return 1
    else:
        return 0


@handle_errors
def fix_gauge_pairwise_model(sequenceLength,
                             alphabetSize,
                             theta):
    """
    Calculates the hierarchical gauge that prioritizes the
    constant feature over pairwise features and the pairwise
    features' coefficients are in the canonical gauge.

     parameters
    ----------
    sequenceLength: (int)
        length of one input sequence

    alphabetSize: (int)
        number of unique bases or amino-acids; e.g. 4 for DNA and 20 for proteins

    theta: (array-like)
        the parameters of the latent neigbhor model

    returns
    -------
    thetaGaugeFixed: (array-like)
        the gauge fixed pairwise parameters

    centeringConstant: (array-like)
        the additive shift relating phi to phi''

     """

    numPosPairs = int(sequenceLength * (sequenceLength - 1) / 2)
    theta = theta.reshape(numPosPairs, alphabetSize, alphabetSize)

    # center parameter values corresponding to a fixed first index, i.e. fixed position pair:
    centeringConstant = sum(theta.mean(axis=(1, 2)))
    thetaCentered = theta - theta.mean(axis=(1, 2), keepdims=True)

    # calculate canonical gauge for pairwise features:
    thetaGaugeFixed = thetaCentered
    if sequenceLength > 2:
        sRow = np.sum(thetaCentered, axis=2,
                      keepdims=True)  # define array that broadcasts with theta and contains theta's row sums
        sCol = np.sum(thetaCentered, axis=1,
                      keepdims=True)  # define array that broadcasts with theta and contains theta's column sums

        # reshape thetaGaugeFixed, sRow, and sCol so that they are inexed as [pos1][pos2,char1,char2] where pos1<pos2
        l = [i for i in range(1, sequenceLength)]
        splittingPoints = [sum(l[-k:]) for k in range(1,
                                                      sequenceLength)]  # [(sequenceLength-1), (sequenceLength-1+sequenceLength-2),...,(sequenceLength-1+sequenceLength-2+...+1)]
        thetaGaugeFixedReshaped = np.array(np.split(thetaGaugeFixed, splittingPoints, axis=0), dtype=object)
        sRowReshaped = np.array(np.split(sRow, splittingPoints, axis=0), dtype=object)
        sColReshaped = np.array(np.split(sCol, splittingPoints, axis=0), dtype=object)

        c = 1 / (alphabetSize * (sequenceLength - 1))
        for pos1 in range(sequenceLength - 1):
            for pos2 in range(pos1 + 1, sequenceLength):
                for pos in range(pos1):
                    thetaGaugeFixedReshaped[pos1][pos2 - (pos1 + 1)] += c * (-(sequenceLength - 2)) ** kronecker_delta(
                        pos, pos2) * sColReshaped[pos][pos1 - (pos + 1)].transpose()
                for pos in range(pos1 + 1, sequenceLength):
                    thetaGaugeFixedReshaped[pos1][pos2 - (pos1 + 1)] += c * (-(sequenceLength - 2)) ** kronecker_delta(
                        pos, pos2) * sRowReshaped[pos1][pos - (pos1 + 1)]
                for pos in range(pos2):
                    thetaGaugeFixedReshaped[pos1][pos2 - (pos1 + 1)] += c * (-(sequenceLength - 2)) ** kronecker_delta(
                        pos, pos1) * sColReshaped[pos][pos2 - (pos + 1)]
                for pos in range(pos2 + 1, sequenceLength):
                    thetaGaugeFixedReshaped[pos1][pos2 - (pos1 + 1)] += c * (-(sequenceLength - 2)) ** kronecker_delta(
                        pos, pos1) * sRowReshaped[pos2][pos - (pos2 + 1)].transpose()
    thetaGaugeFixed = thetaGaugeFixed.ravel()

    return thetaGaugeFixed, centeringConstant





import mavenn
import numpy as np
from mavenn.src.utils import na_plots_for_mavenn_demo, get_example_dataset

# load data
sequences, bin_counts = get_example_dataset(name='Sort-Seq')

# load mavenn's NA model
NAR = mavenn.Model(regression_type='NA',
                   X=sequences,
                   y=bin_counts,
                   model_type='additive',
                   alphabet_dict='dna',
                   ohe_single_batch_size=50000)

NAR.fit(epochs=200,
        use_early_stopping=True,
        early_stopping_patience=20,
        verbose=1)

loss_history =  NAR.model.return_loss()

# evaluate the inferred noise model
#noise_model = NAR.na_noisemodel(sequences)
noise_model, phi_range, latent_trait = NAR.na_noisemodel(sequences, gauge_fix=True)

# plot results using helper function
na_plots_for_mavenn_demo(loss_history, NAR, noise_model, phi_range)



from __future__ import division
import pandas as pd
from mavenn.src.error_handling import check, handle_errors


@handle_errors
def validate_input(df):
    """
    Checks to make sure that the input dataframe, df, contains
    sequences and values that are valid for mavenn. Sequences
    must all be of the same length and values have to be
    integers or floats and not nan or string etc.

    parameters
    ----------

    df: (dataframe)
        A pandas dataframe containing two columns: (i) sequences and (ii)
        values.

    returns
    -------
    out_df: (dataframe)
        A cleaned-up version of df (if possible).
    """

    # check that df is a valid dataframe
    check(isinstance(df, pd.DataFrame),
          'Input data needs to be a valid pandas dataframe, ' 
          'input entered: %s' % type(df))

    # create copy of df so we don't overwrite the user's data
    out_df = df.copy()

    # make sure the input df has only sequences and values
    # and no additional columns
    check(out_df.shape[1] == 2, 'Input dataframe must only have 2 columns, '
                                'sequences and values. Entered # columns %d' % len(out_df.columns))

    # check that 'sequences' and 'values columns are part of the df columns
    check('sequence' in out_df.columns, 'Column containing sequences must be named "sequence" ')
    check('values' in out_df.columns, 'Column containing values must be named "values" ')

    # TODO: check that sequence column is of type string and values column is float or int
    # TODO: need to check that sequences are of the same length, but this could take a lot of time checking

    # return cleaned-up out_df
    out_df = out_df.dropna().copy()
    return out_df





from __future__ import print_function   # so that print behaves like python 3.x not a special lambda statement

import sys
sys.path.insert(0,'../../')
import mavenn
print(mavenn.__path__)


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

global_success_counter = 0
global_fail_counter = 0

# Common success and fail lists
bool_fail_list = [0, -1, 'True', 'x', 1]
bool_success_list = [False, True]


# helper method for functional test_for_mistake
def test_for_mistake(func, *args, **kw):
    """
    Run a function with the specified parameters and register whether
    success or failure was a mistake

    parameters
    ----------

    func: (function or class constructor)
        An executable function to which *args and **kwargs are passed.

    return
    ------

    None.
    """

    global global_fail_counter
    global global_success_counter

    # print test number
    test_num = global_fail_counter + global_success_counter
    print('Test # %d: ' % test_num, end='')
    #print('Test # %d: ' % test_num)

    # Run function
    obj = func(*args, **kw)
    # Increment appropriate counter
    if obj.mistake:
        global_fail_counter += 1
    else:
        global_success_counter += 1


def test_parameter_values(func,
                          var_name=None,
                          fail_list=[],
                          success_list=[],
                          **kwargs):
    """
    Tests predictable success & failure of different values for a
    specified parameter when passed to a specified function

    parameters
    ----------

    func: (function)
        Executable to test. Can be function or class constructor.

    var_name: (str)
        Name of variable to test. If not specified, function is
        tested for success in the absence of any passed parameters.

    fail_list: (list)
        List of values for specified variable that should fail

    success_list: (list)
        List of values for specified variable that should succeed

    **kwargs:
        Other keyword variables to pass onto func.

    return
    ------

    None.

    """

    # If variable name is specified, test each value in fail_list
    # and success_list
    if var_name is not None:

        # User feedback
        print("Testing %s() parameter %s ..." % (func.__name__, var_name))

        # Test parameter values that should fail
        for x in fail_list:
            kwargs[var_name] = x
            test_for_mistake(func=func, should_fail=True, **kwargs)

        # Test parameter values that should succeed
        for x in success_list:
            kwargs[var_name] = x
            test_for_mistake(func=func, should_fail=False, **kwargs)

        print("Tests passed: %d. Tests failed: %d.\n" %
              (global_success_counter, global_fail_counter))

    # Otherwise, make sure function without parameters succeeds
    else:

        # User feedback
        print("Testing %s() without parameters." % func.__name__)

        # Test function
        test_for_mistake(func=func, should_fail=False, **kwargs)

    # close all figures that might have been generated
    plt.close('all')


def test_GlobalEpistasisModel():

    # df inputs that successfully execute when entered into mavenn
    good_input_df = pd.read_csv('../../raw_data/mpsa/psi_9nt_mavenn.csv')
    # df inputs that fail when entered into mavenn.
    bad_df1 = 'x'
    bad_input_df2 = pd.read_csv('../../raw_data/mpsa/psi_9nt.csv')

    print(dir(mavenn))

    # test parameter df
    test_parameter_values(func=mavenn.GlobalEpistasisModel, var_name='df',fail_list=[bool_fail_list,bad_df1
        ,bad_input_df2], success_list=[good_input_df],model_type='additive')


test_GlobalEpistasisModel()


# Classes / functions imported with mavenn
from mavenn.src.model import Model
from mavenn.src.UI import GlobalEpistasisModel
from mavenn.src.UI import NoiseAgnosticModel
from mavenn.src.error_handling import handle_errors, check

# imports required for helper functions in demos
import pandas as pd
import mavenn
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

import os
import re


@handle_errors
def demo(name='GEmpsa'):

    """
    Performs a demonstration of the mavenn software.

    parameters
    -----------

    name: (str)
        Must be one of {'GEmpsa, GEGB1, GEmpsaPairwise, NAsortseq'}.

    returns
    -------
    None.

    """

    # build list of demo names and corresponding file names
    example_dir = '%s/examples' % os.path.dirname(__file__)
    all_base_file_names = os.listdir(example_dir)

    example_file_names = ['%s/%s' % (example_dir, temp_name)
                     for temp_name in all_base_file_names
                     if re.match('demo_.*\.py', temp_name)]

    examples_dict = {}
    for file_name in example_file_names:
        key = file_name.split('_')[-1][:-3]
        examples_dict[key] = file_name

    # check that name is valid
    check(name in examples_dict.keys(),
          'name = %s is not valid. Must be one of %s'
          % (repr(name), examples_dict.keys()))

    # open and run example file
    file_name = examples_dict[name]
    with open(file_name, 'r') as f:
        content = f.read()
        line = '-------------------------------------------------------------'
        print('Running %s:\n%s\n%s\n%s' % \
              (file_name, line, content, line))
    exec(open(file_name).read())




from mavenn.src.error_handling import handle_errors, check
from mavenn.src.UI import GlobalEpistasisModel, NoiseAgnosticModel


@handle_errors
class Model:

    """
    Mavenn's model class that lets the user choose either
    global epistasis regression or noise agnostic regression


    attributes
    ----------

    regression_type: (str)
        variable that choose type of regression, valid options
        include 'GE', 'NA'

    X: (array-like)
        Input pandas DataFrame containing sequences. X are
        DNA, RNA, or protein sequences to be regressed over

    y: (array-like)
        y represents counts in bins corresponding to the sequences X

    model_type: (str)
        Model type specifies the type of GE model the user wants to run.
        Three possible choices allowed: ['additive','neighbor','all-pairs']

    learning_rate: (float)
            Learning rate of the optimizer.

    test_size: (float in (0,1))
        Fraction of data to be set aside as unseen test data for model evaluation
        error.

    monotonic: (boolean)
        Indicates whether to use monotonicity constraint in GE nonlinear function
        or NA noise model. If true then weights of GE nonlinear function, or NA
        noise model, will be constraned to be non-negative.

    alphabet_dict: (str)
        Specifies the type of input sequences. Three possible choices
        allowed: ['dna','rna','protein']

    custom_architecture: (tf.model)
        a custom neural network architecture that replaces the
        default architecture implemented.

    ohe_single_batch_size: (int)
        integer specifying how many sequences to one-hot encode at a time.
        The larger this number number, the quicker the encoding will happen,
        but this may also take up a lot of memory and throw an exception
        if its too large. Currently for additive models only.


    """

    def __init__(self,
                 regression_type,
                 X,
                 y,
                 model_type,
                 learning_rate=0.005,
                 test_size=0.2,
                 monotonic=True,
                 alphabet_dict='dna',
                 custom_architecture=None,
                 ohe_single_batch_size=10000):

        # set class attributes
        self.regression_type = regression_type
        self.X, self.y = X, y
        self.model_type = model_type
        self.learning_rate = learning_rate
        self.test_size = test_size
        self.monotonic = monotonic
        self.alphabet_dict = alphabet_dict
        self.custom_architecture = custom_architecture
        self.ohe_single_batch_size = ohe_single_batch_size

        # class attributes that are not parameters
        # this attribute will instantiate either
        # NA class or the GE class
        self.model = None

        # check that regression_type is valid
        check(self.regression_type in {'NA', 'GE'},
              'regression_type = %s; must be "NA", or  "GE"' %
              self.model_type)

        # choose model based on regression_type
        if regression_type == 'GE':
            self.model = GlobalEpistasisModel(X=self.X,
                                              y=self.y,
                                              model_type=self.model_type,
                                              test_size=self.test_size,
                                              alphabet_dict=self.alphabet_dict,
                                              ohe_single_batch_size=self.ohe_single_batch_size)

            self.define_model = self.model.define_model(monotonic=self.monotonic,
                                                        custom_architecture=self.custom_architecture)

            self.model.compile_model(lr=self.learning_rate)

        elif regression_type == 'NA':
            self.model = NoiseAgnosticModel(self.X,
                                            self.y,
                                            self.model_type,
                                            self.test_size,
                                            self.alphabet_dict,
                                            self.custom_architecture,
                                            self.ohe_single_batch_size)

            self.define_model = self.model.define_model(monotonic=self.monotonic,
                                                        custom_architecture=self.custom_architecture)

            self.model.compile_model(lr=self.learning_rate)

    @handle_errors
    def fit(self,
            validation_split=0.2,
            epochs=50,
            verbose=1,
            use_early_stopping=True,
            early_stopping_patience=50):

        """

        Method that will fit the noise agnostic model to data.

        parameters
        ----------
        validation_split: (float in [0,1])
            Fraction of training data to be split into a validation set.

        epochs: (int>0)
            Number of epochs to complete during training.

        verbose: (0 or 1, or boolean)
            Boolean variable that will show training progress if 1 or True, nothing
            if 0 or False.

        use_early_stopping: (bool)
            specifies whether to use early stopping or not

        early_stopping_patience: (int)
            number of epochs to wait before executing early stopping.


        returns
        -------
        model: (tf.model)
            Trained model that can now be evaluated on the test set
            and used for predictions.

        """

        history = self.model.fit(epochs=epochs,
                                 verbose=verbose,
                                 validation_split=validation_split,
                                 use_early_stopping=use_early_stopping,
                                 early_stopping_patience=early_stopping_patience)
        return history

    @handle_errors
    def ge_nonlinearity(self,
                        sequences,
                        input_range=None,
                        gauge_fix=True):

        """
        Method used to plot GE non-linearity.

        parameters
        ----------

        sequences: (array-like)
            sequences for which the additive trait will be computed.

        input_range: (array-like)
            data range which will be input to the GE nonlinearity.  If
            this is none than range will be determined from min and max
            of the latent trait.

        gauge_fix: (bool)
            if true parameters used to compute latent trait will be
            gauge fixed

        returns
        -------
        ge_nonlinearfucntion: (array-like function)
            the nonlinear GE function.

        """

        check(self.regression_type == 'GE', 'regression type must be "GE" for this function ')

        if input_range is None:
            ge_nonlinearfunction, input_range, latent_trait = self.model.ge_nonlinearity(sequences,
                                                                                         input_range=None,
                                                                                         gauge_fix=gauge_fix)
            # if input range not provided, return input range
            # so ge nonliearity can be plotted against it.
            return ge_nonlinearfunction, input_range, latent_trait
        else:
            ge_nonlinearfunction = self.model.ge_nonlinearity(sequences,
                                                              input_range=input_range,
                                                              gauge_fix=gauge_fix)
            return ge_nonlinearfunction



    @handle_errors
    def return_theta(self):

        """
        method that returns a flattened list of gauge-fixed latent model parameters.

        returns
        -------
        theta_gf: (array-like)
            gauge fixed latent trait models
        """

        return self.model.latent_model_parameters()

    @handle_errors
    def na_noisemodel(self,
                      sequences,
                      input_range=None,
                      gauge_fix=True):

        """
        Method used to plot GE non-linearity.

        parameters
        ----------

        sequences: (array-like)
            sequences for which the additive trait will be computed.

        input_range: (array-like)
            data range which will be input to the NA noise model.  If
            this is none than range will be determined from min and max
            of the latent trait.

        gauge_fix: (bool)
            if true parameters used to compute latent trait will be
            gauge fixed

        returns
        -------
        nanoisemodel: (array-like function)
            the noise model inferred from NA regression.

        """

        check(self.regression_type == 'NA', 'regression type must be "GE" for this function ')

        if input_range is None:
            nanoisemodel, input_range, latent_trait = self.model.noise_model(sequences,
                                                                             input_range=None,
                                                                             gauge_fix=gauge_fix)
            # if input range not provided, return input range
            # so noise model can be plotted against it.
            return nanoisemodel, input_range, latent_trait
        else:
            nanoisemodel = self.model.noise_model(sequences,
                                                  input_range=input_range,
                                                  gauge_fix=gauge_fix)
            return nanoisemodel

    @handle_errors
    def nn_model(self):

        """
        method that returns the tf neural network
        from weights can be accessed.
        """

        return self.model.return_model()



from mavenn.src.validate import validate_input
from mavenn.src.error_handling import handle_errors, check
from mavenn.src.utils import onehot_encode_array, \
    _generate_nbr_features_from_sequences, _generate_all_pair_features_from_sequences

from mavenn.src.utils import _center_matrix, fix_gauge_neighbor_model, fix_gauge_pairwise_model

import numpy as np
import tensorflow as tf
import pandas as pd

import tensorflow.keras
from tensorflow.keras.models import Sequential, Model
from tensorflow.keras.layers import Dense, Dropout, Activation, Input, Lambda, Concatenate
from tensorflow.keras.constraints import non_neg as nonneg
from tensorflow.keras.optimizers import Adam
from tensorflow.keras import metrics
from tensorflow.keras import regularizers
from tensorflow.keras import callbacks
import tensorflow.keras.backend as K

import matplotlib.pyplot as plt

"""
NOTE: could put methods that are common to both classes in
in a utils.py module, e.g. _generate_all_pair_features_from_sequences()

Optional tasks
1. Could add a method to display additive models as sequences logos.
2. Could add mutual information approximator for trained NAR model.
"""


@handle_errors
class GlobalEpistasisModel:    

    """
    Class that implements global epistasis regression.


    attributes
    ----------

    X: (array-like)
        Input pandas DataFrame containing sequences. X are
        DNA, RNA, or protein sequences to be regressed over

    y: (array-like)
        y represents counts in bins corresponding to the sequences X

    model_type: (str)
        Model type specifies the type of GE model the user wants to run.
        Three possible choices allowed: ['additive','neighbor','all-pairs']

    test_size: (float in (0,1))
        Fraction of data to be set aside as unseen test data for model evaluation
        error.

    alphabet_dict: (str)
        Specifies the type of input sequences. Three possible choices
        allowed: ['dna','rna','protein'].

    custom_architecture: (tf.model)
        a custom neural network architecture that replaces the
        default architecture implemented.

    ohe_single_batch_size: (int)
        integer specifying how many sequences to one-hot encode at a time.
        The larger this number number, the quicker the encoding will happen,
        but this may also take up a lot of memory and throw an exception
        if its too large. Currently for additive models only.

    """

    def __init__(self,
                 X,
                 y,
                 model_type='additive',
                 test_size=0.2,
                 alphabet_dict='dna',
                 custom_architecture=None,
                 ohe_single_batch_size=10000):

        # set class attributes
        self.X, self.y = X, y
        self.model_type = model_type
        self.test_size = test_size
        self.alphabet_dict = alphabet_dict
        self.custom_architecture = custom_architecture
        self.ohe_single_batch_size = ohe_single_batch_size

        # class attributes that are not parameters
        # but are useful for using trained models
        self.history = None
        self.model = None

        # perform input checks to validate attributes
        self._input_checks()

        self.x_train, self.y_train = self.X, self.y

        #print('One-hot encoding...')
        # # onehot-encode sequences
        # self.input_seqs_ohe = []
        # for _ in range(len(self.x_train)):
        #     self.input_seqs_ohe.append(onehot_sequence(self.x_train[_]).ravel())
        #
        # # turn lists into np arrays for consumption by tf
        # self.input_seqs_ohe = np.array(self.input_seqs_ohe)

        if self.alphabet_dict == 'dna':
            self.bases = ['A', 'C', 'G', 'T']
        elif self.alphabet_dict == 'rna':
            self.bases = ['A', 'C', 'G', 'U']
        elif self.alphabet_dict == 'protein':

            # this should be called amino-acids
            # need to figure out way to deal with
            # naming without changing a bunch of
            # unnecessary refactoring.
            self.bases = ['A', 'C', 'D', 'E', 'F',
                          'G', 'H', 'I', 'K', 'L',
                          'M', 'N', 'P', 'Q', 'R',
                          'S', 'T', 'V', 'W', 'Y']

        if model_type == 'additive':
            # one-hot encode sequences in batches in a vectorized way
            self.input_seqs_ohe = onehot_encode_array(self.x_train, self.bases, self.ohe_single_batch_size)

        elif model_type == 'neighbor':
            # one-hot encode sequences in batches in a vectorized way
            self.input_seqs_ohe = _generate_nbr_features_from_sequences(self.x_train, self.alphabet_dict)

        elif model_type == 'pairwise':
            # one-hot encode sequences in batches in a vectorized way
            self.input_seqs_ohe = _generate_all_pair_features_from_sequences(self.x_train, self.alphabet_dict)

        # check if this is strictly required by tf
        self.y_train = np.array(self.y_train).reshape(np.shape(self.y_train)[0], 1)

    def _input_checks(self):

        """
        Validate parameters passed to the GlobalEpistasis constructor
        """
        # validate input df
        #self.df = validate_input(self.df)

        check(isinstance(self.X, (list, np.ndarray)),
              'type(X) = %s must be of type list or np.array' % type(self.X))

        check(isinstance(self.y, (list, np.ndarray)),
              'type(y) = %s must be of type list or np.array' % type(self.y))

        check(len(self.X) == len(self.y),
              'length of inputs (X, y) must be equal')

        # check that model type valid
        check(self.model_type in {'additive', 'neighbor', 'pairwise'},
              'model_type = %s; must be "additive", "neighbor", or "pairwise"' %
              self.model_type)

    # JBK: User should be able to set these parameters in the constructor
    def define_model(self,
                     monotonic=True,
                     custom_architecture=None):

        """
        Method that will define the architecture of the global epistasis model.
        using the tensorflow.keras functional api. If the subnetwork architecture
        is not None, than archicture is constructed via the self.custom_architecture
        keyword

        parameters
        ----------

        monotonic: (boolean)
            Indicates whether to use monotonicity constraint in GE nonlinear function.
            If true then weights of GE nonlinear function will be constraned to
            be non-negative.

        custom_architecture: (tf.model)
            a custom neural network architecture that replaces the
            default architecture implemented.

        returns
        -------

        model: (tf.model)
            A tensorflow model that can be compiled and subsequently fit to data.


        """

        # user has not provided custom architecture, implement a default architecture
        if custom_architecture is None:

            number_input_layer_nodes = len(self.input_seqs_ohe[0])
            inputTensor = Input((number_input_layer_nodes,), name='Sequence')

            phi = Dense(1, use_bias=True)(inputTensor)

            # implement monotonicity constraints
            if monotonic:
                intermediateTensor = Dense(50, activation='sigmoid', kernel_constraint=nonneg())(phi)
                outputTensor = Dense(1, kernel_constraint=nonneg())(intermediateTensor)
            else:
                intermediateTensor = Dense(50, activation='sigmoid')(phi)
                outputTensor = Dense(1)(intermediateTensor)

            # create the model:

            model = Model(inputTensor, outputTensor)
            self.model = model

            return model

        # if user has provided custom architecture
        else:
            self.model = custom_architecture
            return custom_architecture

    def return_model(self):
        """

        returns (GE model)
            Helper method that returns the model attribute,
            so that it may easily accessible from the mavenn
            Model class
        """

        return self.model

    def compile_model(self,
                      optimizer='Adam',
                      lr=0.00005,
                      metrics=None):

        """
        This method will compile the model created in the define_model method.
        Loss is mean squared error.

        parameters
        ----------
        optimizer: (str)
            Specifies which optimizers to use during training. Valid choices include
            ['Adam', 'SGD', 'RMSPROP', ... link to keras documentation']

        lr: (float)
            Learning rate of the optimizer.

        metrics: (str)
            Optional metrics to track as MSE loss is minimized, e.g. mean absolute
            error. Link to allowed metrics from keras documention...


        returns
        -------
        None


        """
        self.model.compile(loss='mean_squared_error',
                           optimizer=Adam(lr=lr),
                           metrics=['mean_absolute_error'])

    def fit(self,
            validation_split=0.2,
            epochs=50,
            verbose=1,
            use_early_stopping=True,
            early_stopping_patience=50
           ):

        """

        Method that will fit the global epistasis model to data.

        parameters
        ----------
        validation_split: (float in [0,1])
            Fraction of training data to be split into a validation set.

        epochs: (int>0)
            Number of epochs to complete during training.

        verbose: (0 or 1, or boolean)
            Boolean variable that will show training progress if 1 or True, nothing
            if 0 or False.

        use_early_stopping: (bool)
            specifies whether to use early stopping or not

        early_stopping_patience: (int)
            number of epochs to wait before executing early stopping.


        returns
        -------
        model: (tf.model)
            Trained model that can now be evaluated on the test set
            and used for predictions.

        """

        if use_early_stopping:
            esCallBack = tensorflow.keras.callbacks.EarlyStopping(monitor='val_loss',
                                                                  mode='auto',
                                                                  patience=early_stopping_patience)

            history = self.model.fit(self.input_seqs_ohe,
                                     self.y_train,
                                     validation_split=validation_split,
                                     epochs=epochs,
                                     verbose=verbose,
                                     callbacks=[esCallBack]
                                     )

        else:
            history = self.model.fit(self.input_seqs_ohe,
                                     self.y_train,
                                     validation_split=validation_split,
                                     epochs=epochs,
                                     verbose=verbose,
                                     )

        self.history = history
        return history

    # JBK: perhaps just call this "evaluate". Also, why pass "model" argument?
    def model_evaluate(self,
                       model,
                       data):

        """
        Method to evaluate trained model.

        parameters
        ----------

        model: (tf.model)
            A trained tensorflow model.

        data: (array-like)
            Data to be evaluate model on.

        returns
        -------
        Value of loss on the test set.
        """

        pass

    def predict(self,
                data):

        """
        Method to make predictions from trained model

        parameters
        ----------

        data: (array-like)
            Data on which to make predictions.

        returns
        -------

        predictions: (array-like)
            An array of predictions
        """

        # TODO need to do data validation here
        # e.g. check if data is already one-hot encoded

        if self.model_type=='additive':
            # one-hot encode sequences in batches in a vectorized way
            test_input_seqs_ohe = onehot_encode_array(data, self.bases)

        elif self.model_type=='neighbor':
            # one-hot encode sequences in batches in a vectorized way
            test_input_seqs_ohe = _generate_nbr_features_from_sequences(data, self.alphabet_dict)

        elif self.model_type=='pairwise':
            # one-hot encode sequences in batches in a vectorized way
            test_input_seqs_ohe = _generate_all_pair_features_from_sequences(data, self.alphabet_dict)

        return self.model.predict(test_input_seqs_ohe)

    def return_loss(self):

        """
        Method that returns loss values.

        returns
        -------
        history: (object)
            self.attribute/object that contains loss history

        """

        return self.history

    def latent_model_parameters(self):

        """
        method that returns a flattened list of gauge-fixed
        GE latent trait model parameters.

        returns
        -------
        theta_gf: (array-like)
            gauge fixed latent trait models
        """

        # use training sequences to compute phi
        get_1st_layer_output = K.function([self.model.layers[0].input], [self.model.layers[1].output])

        phi_for_computing_norm = get_1st_layer_output([self.input_seqs_ohe])

        # compute the norm using phi_for_computing_norm as:
        norm = np.sqrt(np.var(phi_for_computing_norm[0]))

        # variables required for gauge fixed latent trait parameters theta
        sequence_length = len(self.x_train[0])
        # non-gauge fixed theta
        theta = self.model.layers[1].get_weights()[0]

        # use gauge-fixing method for the appropriate latent model type
        if self.model_type == 'additive':
            theta_df = pd.DataFrame(theta.reshape(sequence_length, len(self.bases)), columns=self.bases)
            # compute mean centered, thus gauge-fixed, additive latent model
            gf_theta, theta_bar = _center_matrix(theta_df)
            gf_theta = gf_theta.values
            # compute offset term which separates latent trait from gauge fixed latent trait
            a = np.sum(self.input_seqs_ohe[0].reshape(sequence_length, len(self.bases)) * theta_bar)
            a = -a

        elif self.model_type == 'neighbor':

            # compute gauge-fixed, neighbor latent model
            gf_theta, a = fix_gauge_neighbor_model(sequence_length, len(self.bases), theta)

        elif self.model_type == 'pairwise':

            # compute gauge-fixed pairwise latent model
            gf_theta, a = fix_gauge_pairwise_model(sequence_length, len(self.bases), theta)

        return gf_theta.ravel() / norm

    def ge_nonlinearity(self,
                        sequences,
                        input_range=None,
                        gauge_fix = True,
                        ):

        """
        Method used to plot GE non-linearity.

        parameters
        ----------

        sequences: (array-like)
            sequences for which the additive trait will be computed.

        input_range: (array-like)
            data range which will be input to the GE nonlinearity.
            If this is none than range will be determined from min
            and max of the latent trait


        returns
        -------
        ge_nonlinearity: (array-like function)
            the nonlinear GE function.

        """
        # TODO input checks on input_range

        # need to do check if linear model, i.e. when no GE network,
        check(len(self.model.layers) > 3, 'cannot make GE nonlinearity with network having < 3 layers')

        # one hot encode sequences and then subsequently use them
        # to compute latent trait
        if self.model_type=='additive':
            # one-hot encode sequences in batches in a vectorized way
            sequences_ohe = onehot_encode_array(sequences, self.bases)

        elif self.model_type=='neighbor':
            # one-hot encode sequences in batches in a vectorized way
            sequences_ohe = _generate_nbr_features_from_sequences(sequences, self.alphabet_dict)

        elif self.model_type=='pairwise':
            # one-hot encode sequences in batches in a vectorized way
            sequences_ohe = _generate_all_pair_features_from_sequences(sequences, self.alphabet_dict)

        get_1st_layer_output = K.function([self.model.layers[0].input], [self.model.layers[1].output])
        latent_trait = get_1st_layer_output([sequences_ohe])

        # tf adds an extra dimensions, so we will remove it
        latent_trait = latent_trait[0].ravel().copy()

        ge_model_input = Input((1,))
        next_input = ge_model_input

        # this variable will contain the index of the latent trait hidden layer
        # and will determine which layers correspond to the GE non-linear subnetwork
        latent_trait_layer_index = 0
        for layer in self.model.layers[::-1]:

            latent_trait_layer_index += 1
            if len(layer.get_weights()[0]) == 1:
                break

        for layer in self.model.layers[-latent_trait_layer_index:]:
        # for layer in self.model.layers[-2:]:
            next_input = layer(next_input)

        # the following model is not gauge fixed
        ge_model = Model(inputs=ge_model_input, outputs=next_input)

        # variables required for gauge fixed latent trait parameters theta
        sequence_length = len(self.x_train[0])
        # non-gauge fixed theta
        theta = self.model.layers[1].get_weights()[0]
        bias = self.model.layers[1].get_weights()[1]

        # use gauge-fixing method for the appropriate latent model type
        if self.model_type == 'additive':
            theta_df = pd.DataFrame(theta.reshape(sequence_length, len(self.bases)), columns=self.bases)
            # compute mean centered, thus gauge-fixed, additive latent model
            gf_theta, theta_bar = _center_matrix(theta_df)
            gf_theta = gf_theta.values
            # compute offset term which separates latent trait from gauge fixed latent trait
            a = np.sum(sequences_ohe[0].reshape(sequence_length, len(self.bases)) * theta_bar)
            a = -a

        elif self.model_type == 'neighbor':

            # compute gauge-fixed, neighbor latent model
            gf_theta, a = fix_gauge_neighbor_model(sequence_length, len(self.bases), theta)

        elif self.model_type == 'pairwise':

            # compute gauge-fixed pairwise latent model
            gf_theta, a = fix_gauge_pairwise_model(sequence_length, len(self.bases), theta)

        # this conditional currently outputs the not gauge-fixed, original, GE nonlinearity
        if gauge_fix is False:
            if input_range is None:
                input_range = np.linspace(min(latent_trait), max(latent_trait), 1000)
                ge_nonlinearity = ge_model.predict(input_range)

                # if input range not provided, return input range
                # so ge nonliearity can be plotted against it.
                return ge_nonlinearity, input_range, latent_trait
            else:
                ge_nonlinearity = ge_model.predict(input_range)
                return ge_nonlinearity

        # this part of the condition returns the gauge fixed latent trait
        elif gauge_fix is True:

            # first compute norm which will be used to rescale the parameters theta
            # compute phi using training sequences
            phi_for_computing_norm = get_1st_layer_output([self.input_seqs_ohe])

            # compute the norm using phi_for_computing_norm as:
            norm = np.sqrt(np.var(phi_for_computing_norm[0]))

            # rescale the gf_theta
            theta_gf_rescaled = gf_theta.ravel() / norm

            # create a copy of the model which will gauge fixed, this
            # way the user can retain non-gauge-fixed model if they want
            gauge_fixed_model = tf.keras.models.clone_model(self.model)

            # update the weights of the latent model with the gauge fixed parameters, and compute gauge fixed phi
            gauge_fixed_model.layers[1].set_weights([theta_gf_rescaled.reshape(len(theta_gf_rescaled), 1),
                                              self.model.layers[1].get_weights()[1]])

            get_gauge_fixed_phi = K.function([gauge_fixed_model.layers[0].input], [gauge_fixed_model.layers[1].output])
            phi_gauge_fixed = get_gauge_fixed_phi([sequences_ohe])

            input_range_gf_phi = np.linspace(min(phi_gauge_fixed[0]), max(phi_gauge_fixed[0]), 1000)

            ge_nonlinearity_gf = ge_model.predict([(input_range_gf_phi-bias)*norm+a+bias])
            return ge_nonlinearity_gf, input_range_gf_phi, phi_gauge_fixed[0]


class NoiseAgnosticModel:

    """
    Class that implements Noise agnostic regression.


    attributes
    ----------

    X: (array-like)
        Input pandas DataFrame containing sequences. X are
        DNA, RNA, or protein sequences to be regressed over

    y: (array-like)
        y represents counts in bins corresponding to the sequences X

    model_type: (str)
        Model type specifies the type of NAR model the user wants to run.
        Three possible choices allowed: ['additive','neighbor','pairwise']

    test_size: (float in (0,1))
        Fraction of data to be set aside as unseen test data for model evaluation
        error.

    alphabet_dict: (str)
        Specifies the type of input sequences. Three possible choices
        allowed: ['dna','rna','protein']

    custom_architecture: (tf.model)
        a custom neural network architecture that replaces the
        default architecture implemented.

    ohe_single_batch_size: (int)
        integer specifying how many sequences to one-hot encode at a time.
        The larger this number number, the quicker the encoding will happen,
        but this may also take up a lot of memory and throw an exception
        if its too large. Currently for additive models only.

    """

    def __init__(self,
                 X,
                 y,
                 model_type='additive',
                 test_size=0.2,
                 alphabet_dict='dna',
                 custom_architecture=None,
                 ohe_single_batch_size=10000):

        # set class attributes
        self.X = X
        self.y = y
        self.model_type = model_type
        self.test_size = test_size
        self.alphabet_dict = alphabet_dict
        self.custom_architecture = custom_architecture
        self.ohe_single_batch_size = ohe_single_batch_size

        # class attributes that are not parameters
        # but are useful for using trained models
        self.history = None
        self.model = None

        # perform input checks to validate attributes
        self._input_checks()

        self.x_train, self.y_train = self.X, self.y

        #print('One-hot encoding...')
        # # onehot-encode sequences
        # self.input_seqs_ohe = []
        # for _ in range(len(self.x_train)):
        #     self.input_seqs_ohe.append(onehot_sequence(self.x_train[_]).ravel())
        #
        # # turn lists into np arrays for consumption by tf
        # self.input_seqs_ohe = np.array(self.input_seqs_ohe)

        if self.alphabet_dict == 'dna':
            self.bases = ['A', 'C', 'G', 'T']
        elif self.alphabet_dict == 'rna':
            self.bases = ['A', 'C', 'G', 'U']
        elif self.alphabet_dict == 'protein':

            # this should be called amino-acids
            # need to figure out way to deal with
            # naming without changing a bunch of
            # unnecessary refactoring.
            self.bases = ['A', 'C', 'D', 'E', 'F',
                          'G', 'H', 'I', 'K', 'L',
                          'M', 'N', 'P', 'Q', 'R',
                          'S', 'T', 'V', 'W', 'Y']

        if model_type=='additive':
            # one-hot encode sequences in batches in a vectorized way
            self.input_seqs_ohe = onehot_encode_array(self.x_train, self.bases, self.ohe_single_batch_size)

        elif model_type=='neighbor':
            # one-hot encode sequences in batches in a vectorized way
            self.input_seqs_ohe = _generate_nbr_features_from_sequences(self.x_train, self.alphabet_dict)

        elif model_type=='pairwise':
            # one-hot encode sequences in batches in a vectorized way
            self.input_seqs_ohe = _generate_all_pair_features_from_sequences(self.x_train, self.alphabet_dict)

        # check if this is strictly required by tf
        self.y_train = np.array(self.y_train)

    def _input_checks(self):

        """
        Validate parameters passed to the NoiseAgnosticRegression constructor
        """
        check(isinstance(self.X, (list, np.ndarray)),
              'type(X) = %s must be of type list or np.array' % type(self.X))

        check(isinstance(self.y, (list, np.ndarray)),
              'type(y) = %s must be of type list or np.array' % type(self.y))

        check(len(self.X) == len(self.y),
              'length of inputs (X, y) must be equal')

        # check that model type valid
        check(self.model_type in {'additive', 'neighbor', 'pairwise'},
              'model_type = %s; must be "additive", "neighbor", or "pairwise"' %
              self.model_type)

    def define_model(self,
                     monotonic=True,
                     custom_architecture=None):

        """
        Method that will define the architecture of the global epistasis model.
        using the tensorflow.keras functional api. If the subnetwork architecture
        is not None, than archicture is constructed via the dict sub_network_layers_nodes_dict

        parameters
        ----------

        monotonic: (boolean)
            If True, than weights in noise model will be constrained to be non-negative.

        custom_architecture: (tf.model)
            a custom neural network architecture that replaces the
            default architecture implemented.

        returns
        -------

        model: (tf.model)
            A tensorflow.keras model that can be compiled and subsequently fit to data.


        """

        if custom_architecture is None:
            number_input_layer_nodes = len(self.input_seqs_ohe[0])
            inputTensor = Input((number_input_layer_nodes,), name='Sequence')

            phi = Dense(1, use_bias=True, name='additive_weights')(inputTensor)

            # implement monotonicity constraints
            if monotonic:
                intermediateTensor = Dense(10, activation='sigmoid', kernel_constraint=nonneg())(phi)
                outputTensor = Dense(np.shape(self.y_train[0])[0], activation='softmax', kernel_constraint=nonneg())(
                    intermediateTensor)
            else:
                intermediateTensor = Dense(10, activation='sigmoid')(phi)
                outputTensor = Dense(np.shape(self.y_train[0])[0], activation='softmax')(intermediateTensor)


            # #create the model:
            model = Model(inputTensor, outputTensor)
            self.model = model
            return model
        else:
            self.model = custom_architecture
            return custom_architecture

    def compile_model(self,
                      lr=0.005):

        """
        This method will compile the model created in the define_model method.
        The loss used will be log_poisson_loss.

        parameters
        ----------

        optimizer: (str)
            Specifies which optimizers to use during training. Valid choices include
            ['Adam', 'SGD', 'RMSPROP', ... link to keras documentation']

        lr: (float)
            Learning rate of the optimizer.

        metrics: (str)
            Optional metrics to track as MSE loss is minimized, e.g. mean absolute
            error. Link to allowed metrics from keras documention...


        returns
        -------
        None

        """

        self.model.compile(loss=tf.nn.log_poisson_loss,
                           optimizer=Adam(lr=lr),
                           metrics=['categorical_accuracy'])

    def fit(self,
            validation_split=0.2,
            epochs=50,
            verbose=1,
            use_early_stopping=True,
            early_stopping_patience=50):

        """

        Method that will fit the noise agnostic model to data.

        parameters
        ----------
        validation_split: (float in [0,1])
            Fraction of training data to be split into a validation set.

        epochs: (int>0)
            Number of epochs to complete during training.

        verbose: (0 or 1, or boolean)
            Boolean variable that will show training progress if 1 or True, nothing
            if 0 or False.

        use_early_stopping: (bool)
            specifies whether to use early stopping or not

        early_stopping_patience: (int)
            number of epochs to wait before executing early stopping.


        returns
        -------
        model: (tf.model)
            Trained model that can now be evaluated on the test set
            and used for predictions.

        """

        if use_early_stopping:
            esCallBack = tensorflow.keras.callbacks.EarlyStopping(monitor='val_loss',
                                                                  mode='auto',
                                                                  patience=early_stopping_patience)

            history = self.model.fit(self.input_seqs_ohe,
                                     self.y_train,
                                     validation_split=validation_split,
                                     epochs=epochs,
                                     verbose=verbose,
                                     callbacks=[esCallBack]
                                     )

        else:
            history = self.model.fit(self.input_seqs_ohe,
                                     self.y_train,
                                     validation_split=validation_split,
                                     epochs=epochs,
                                     verbose=verbose,
                                     )

        self.history = history
        return history

    def model_evaluate(self,
                       model,
                       data):

        """
        Method to evaluate trained model.

        parameters
        ----------

        model: (tf.model)
            A trained tensorflow model.

        data: (array-like)
            Data to be evaluate model on.

        returns
        -------
        Value of loss on the test set.
        """

        pass

    def return_loss(self):

        """
        Method that returns loss values.

        returns
        -------
        history: (object)
            self.attribute/object that contains loss history

        """

        return self.history

    def return_model(self):
        """

        returns (GE model)
            Helper method that returns the model attribute,
            so that it may easily accessible from the mavenn
            Model class
        """

        return self.model


    def latent_model_parameters(self):

        """
        method that returns a flattened list of gauge-fixed
        NA latent trait model parameters.

        returns
        -------
        theta_gf: (array-like)
            gauge fixed latent trait models
        """

        # use training sequences to compute phi
        get_1st_layer_output = K.function([self.model.layers[0].input], [self.model.layers[1].output])

        phi_for_computing_norm = get_1st_layer_output([self.input_seqs_ohe])

        # compute the norm using phi_for_computing_norm as:
        norm = np.sqrt(np.var(phi_for_computing_norm[0]))

        # variables required for gauge fixed latent trait parameters theta
        sequence_length = len(self.x_train[0])
        # non-gauge fixed theta
        theta = self.model.layers[1].get_weights()[0]

        # use gauge-fixing method for the appropriate latent model type
        if self.model_type == 'additive':
            theta_df = pd.DataFrame(theta.reshape(sequence_length, len(self.bases)), columns=self.bases)
            # compute mean centered, thus gauge-fixed, additive latent model
            gf_theta, theta_bar = _center_matrix(theta_df)
            gf_theta = gf_theta.values
            # compute offset term which separates latent trait from gauge fixed latent trait
            a = np.sum(self.input_seqs_ohe[0].reshape(sequence_length, len(self.bases)) * theta_bar)
            a = -a

        elif self.model_type == 'neighbor':

            # compute gauge-fixed, neighbor latent model
            gf_theta, a = fix_gauge_neighbor_model(sequence_length, len(self.bases), theta)

        elif self.model_type == 'pairwise':

            # compute gauge-fixed pairwise latent model
            gf_theta, a = fix_gauge_pairwise_model(sequence_length, len(self.bases), theta)

        return gf_theta.ravel() / norm

    def noise_model(self,
                    sequences=None,
                    input_range=None,
                    gauge_fix=True):

        """
        Method used to compute NA noise model.

        parameters
        ----------

        sequences: (array-like)
            sequences for which the additive trait will be computed.

        input_range: (array-like)
            data range which will be input to the noise model.
            If this is none than range will be determined from min
            and max of the latent trait

        gauge_fix: (bool)
            if true parameters used to compute latent trait will be
            gauge fixed

        returns
        -------
        noise_model: (array-like function)
            the noise model mapping latent trait to bins.

        """

        # TODO input checks on input_range and sequences

        # one hot encode sequences and then subsequently use them
        # to compute latent trait

        # compute latent trait if sequences given
        if sequences is not None:

            if self.model_type=='additive':
                # one-hot encode sequences in batches in a vectorized way
                sequences_ohe = onehot_encode_array(sequences, self.bases)

            elif self.model_type=='neighbor':
                # one-hot encode sequences in batches in a vectorized way
                sequences_ohe = _generate_nbr_features_from_sequences(sequences, self.alphabet_dict)

            elif self.model_type=='pairwise':
                # one-hot encode sequences in batches in a vectorized way
                sequences_ohe = _generate_all_pair_features_from_sequences(sequences, self.alphabet_dict)

            get_1st_layer_output = K.function([self.model.layers[0].input], [self.model.layers[1].output])
            latent_trait = get_1st_layer_output([sequences_ohe])

            # tf adds an extra dimensions, so we will remove it
            latent_trait = latent_trait[0].ravel().copy()

        noise_model_input = Input((1,))
        next_input = noise_model_input

        # this variable will contain the index of the latent trait hidden layer
        # and will determine which layers correspond to the GE non-linear subnetwork
        latent_trait_layer_index = 0
        for layer in self.model.layers[::-1]:

            latent_trait_layer_index += 1
            if len(layer.get_weights()[0]) == 1:
                break

        for layer in self.model.layers[-latent_trait_layer_index:]:
        #for layer in self.model.layers[-2:]:
            next_input = layer(next_input)

        noise_model = Model(inputs=noise_model_input, outputs=next_input)

        # variables required for gauge fixed latent trait parameters theta
        sequence_length = len(self.x_train[0])
        # non-gauge fixed theta
        theta = self.model.layers[1].get_weights()[0]
        bias = self.model.layers[1].get_weights()[1]

        # use gauge-fixing method for the appropriate latent model type
        if self.model_type == 'additive':
            theta_df = pd.DataFrame(theta.reshape(sequence_length, len(self.bases)), columns=self.bases)
            # compute mean centered, thus gauge-fixed, additive latent model
            gf_theta, theta_bar = _center_matrix(theta_df)
            gf_theta = gf_theta.values
            # compute offset term which separates latent trait from gauge fixed latent trait
            a = np.sum(sequences_ohe[0].reshape(sequence_length, len(self.bases)) * theta_bar)
            a = -a

        elif self.model_type == 'neighbor':

            # compute gauge-fixed, neighbor latent model
            gf_theta, a = fix_gauge_neighbor_model(sequence_length, len(self.bases), theta)

        elif self.model_type == 'pairwise':

            # compute gauge-fixed pairwise latent model
            gf_theta, a = fix_gauge_pairwise_model(sequence_length, len(self.bases), theta)

        # this conditional currently outputs the not gauge-fixed, original, GE nonlinearity
        if gauge_fix is False:
            if input_range is None and sequences is not None:
                input_range = np.linspace(min(latent_trait), max(latent_trait), 1000)
                noise_function = noise_model.predict(input_range)

                # if input range not provided, return input range
                # so ge nonliearity can be plotted against it.
                return noise_function, input_range, latent_trait
            else:
                noise_function = noise_model.predict(input_range)
                return noise_function

        # this part of the condition returns the gauge fixed latent trait
        elif gauge_fix is True:

            # first compute norm which will be used to rescale the parameters theta
            # compute phi using training sequences
            phi_for_computing_norm = get_1st_layer_output([self.input_seqs_ohe])

            # compute the norm using phi_for_computing_norm as:
            norm = np.sqrt(np.var(phi_for_computing_norm[0]))

            # rescale the gf_theta
            theta_gf_rescaled = gf_theta.ravel() / norm

            # create a copy of the model which will gauge fixed, this
            # way the user can retain non-gauge-fixed model if they want
            gauge_fixed_model = tf.keras.models.clone_model(self.model)

            # update the weights of the latent model with the gauge fixed parameters, and compute gauge fixed phi
            gauge_fixed_model.layers[1].set_weights([theta_gf_rescaled.reshape(len(theta_gf_rescaled), 1),
                                              self.model.layers[1].get_weights()[1]])

            # update the weights of the latent model with the gauge fixed parameters, and compute gauge fixed phi
            gauge_fixed_model.layers[1].set_weights([theta_gf_rescaled.reshape(len(theta_gf_rescaled), 1),
                                              self.model.layers[1].get_weights()[1]])

            get_gauge_fixed_phi = K.function([gauge_fixed_model.layers[0].input], [gauge_fixed_model.layers[1].output])
            phi_gauge_fixed = get_gauge_fixed_phi([sequences_ohe])

            input_range_gf_phi = np.linspace(min(phi_gauge_fixed[0]), max(phi_gauge_fixed[0]), 1000)

            noise_model = noise_model.predict([(input_range_gf_phi-bias)*norm+a+bias])
            return noise_model, input_range_gf_phi, phi_gauge_fixed[0]



import mavenn
import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split
from mavenn.src.utils import ge_plots_for_mavenn_demo, get_example_dataset

# load data
X, y = get_example_dataset(name='MPSA')

# split data into training and testing sets
x_train, x_test, y_train, y_test = train_test_split(X, y)

# load mavenn's GE model
GER = mavenn.Model(regression_type='GE',
                   X=x_train,
                   y=y_train,
                   model_type='additive',
                   learning_rate=0.005,
                   alphabet_dict='rna')

GER.fit(epochs=200, use_early_stopping=True, early_stopping_patience=20, verbose=1)

# make predictions on held out test set
loss_history = GER.model.return_loss()
predictions = GER.model.predict(x_test)

# plot results using helper function
ge_plots_for_mavenn_demo(loss_history, predictions, y_test, x_test, GER)





from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()

setup(name='mavenn',
      version='0.14',
      description='package for inferring models of sequence-function relationships from mave datasets',
      long_description=readme(),
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
      ],
      keywords='mave, mpa',
      url='http://mavenn.readthedocs.io',
      author='Ammar Tareen, Justin B. Kinney',
      author_email='tareen@cshl.edu',
      license='MIT',
      packages=['mavenn'],
      include_package_data=True,
      install_requires=[
        'numpy',
		'matplotlib',
		'pandas',
		'tensorflow',
		'sklearn',
		'logomaker',
		'seaborn'
      ],
      zip_safe=False)